unit NSysW;

interface

uses Windows, NTypes, NLibW;

type
  TTime = record
    Year: Integer;
    Mon: 1..12;
    MDay: 1..31;
    Hour: 0..23;
    Min: 0..59;
    Sec: 0..61;
    WDay: 0..6;
    UTCOffset: Integer;
  end;

  TFileMode = set of (fmRead, fmWrite, fmReadWrite, fmTruncate);

  TFile = class(TStream)
  protected
    FHandle: THandle;
    FName: WideString;
    IsStdHandle: Boolean;
    procedure CreateHandle(const Name: WideString; Mode: TFileMode; Create_: DWORD);
  public
    constructor Create(const Name: WideString; Mode: TFileMode=[fmWrite]);
    constructor Open(const Name: WideString; Mode: TFileMode=[fmRead]);
    constructor CreateStdOut;
    constructor CreateStdIn;
    constructor CreateStdErr;
    constructor OpenConOut;
    constructor OpenConIn;
    constructor CreateByHandle(Handle: THandle);
    destructor Destroy; override;
    function Read(var Buf; NBytes: Cardinal): Cardinal; override;
    function Write(const Buf; NBytes: Cardinal): Cardinal; override;
    function Seek(Offset: Int64; Whence: TSeekWhence=swSet): Int64; override;
    function Seekable: Boolean; override;
    function Size: Cardinal; override;
    function IsConsole: Boolean;
    property Handle: THandle read FHandle;
    property Name: WideString read FName;
  end;

  TDirMode = set of (dmDirectory, dmReadByOwner, dmWriteByOwner);

  TDir = record
    Name: WideString;
    Size: Int64;
    ATime: Int64;
    MTime: Int64;
    Mode: TDirMode;
  end;

  TListDir = class(TObject)
  protected
    Handle: THandle;
    FindData: TWin32FindDataW;
    First: Boolean;
  public
    Path: WideString;
    constructor Create(const Path: WideString);
    destructor Destroy; override;
    function Each(var Dir: TDir): Boolean;
  end;

  TProcess = class(TObject)
  protected
    ProcessInformation: TProcessInformation;
  public
    StdOut: TFile;
    constructor Create(const Args: array of WideString; const StdIn: TFile;
        const StdOut: TFile = nil);
    destructor Destroy; override;
    function Wait: Integer;
    procedure Times(var System, User: Int64);
  end;

procedure KillProcess(Pid: Longint);

procedure Remove(const Path: string);
procedure CreateDir(const Path: string);
procedure ChDir(const Path: string);
function GetWd: string;

function SysErrNo: Integer;
function SysError(Code: Integer): string;

procedure SetEnv(const Name: string; const Value: string);
function GetEnv(const Name: WideString): WideString;
function GetEnviron: TStringArray;

function Exists(const Path: WideString): Boolean;
function IsDir(const Path: WideString): Boolean;
function IsFile(const Path: WideString): Boolean;
function DirStat(const Path: WideString): TDir;
procedure DirFWStat(const Path: WideString; const Dir: TDir);
procedure DirWStat(const F: TFile; const Dir: TDir);
procedure ClearDir(var Dir: TDir);
function SameFile(const Path1: WideString; const Path2: WideString): Boolean;

function Time: Int64;
function NanoSec: Int64;
procedure SetNanoSec(const NanoSec: Int64);
function UTCTime(const T: Int64): TTime;
function LocalTime(const T: Int64): TTime;
procedure Sleep(MilliSecs: Cardinal);
function Clock: Cardinal;

procedure SplitDrive(const Path: WideString; var Drive, Tail: WideString);
procedure SplitExt(const Path: WideString; var Root, Ext: WideString);
procedure SplitPath(const Path: WideString; var Head, Tail: WideString);
function JoinPath(const Paths: array of WideString): WideString;
function AbsPath(const Path: WideString): WideString;
function IsAbsPath(const Path: WideString): Boolean;

function StrColl(const S1, S2: string): Integer;
function StrCollW(const S1, S2: WideString): Integer;
function StrUpper(const S: string): string;

function GenRandom(Len: Longint): string;

const
  EOL = #13#10;
  DevNull = 'NUL';

var
  Argv: TWideStringArray;
  ProgName: WideString;
  ProgPath: WideString;
  StdIn, StdOut, StdErr: TFile;

implementation

uses WinCryptSub;

const
  PathSep = '\';
  AltPathSep = '/';
  DriveSep = ':';
  ExtSep = '.';

  UnixTimeDiff: Int64 = 116444736000000000;

{ private functions }

function FindLastSep(const Path: WideString; Sep: WideChar; AltSep: WideChar): Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := Length(Path) downto 1 do
    if (Path[i] = Sep) or (Path[i] = AltSep) then
    begin
      Result := i;
      Break;
    end;
end;

function FileTimeToUnixTime(const T: TFileTime): Int64;
begin
  Result := (Int64(T) - UnixTimeDiff) div 10000000;
end;

function UnixTimeToFileTime(const T: Int64): TFileTime;
begin
  Int64(Result) := T * 10000000 + UnixTimeDiff;
end;

function FindDataToDir(const FindData: TWin32FindDataW): TDir;
begin
  with FindData do
  begin
    Result.Name := cFileName;
    Result.Size := Int64(nFileSizeHigh) shl 32 + nFileSizeLow;
    Result.Mode := [dmReadByOwner, dmWriteByOwner];
    if (dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0 then
      Include(Result.Mode, dmDirectory);
    if (dwFileAttributes and FILE_ATTRIBUTE_READONLY) <> 0 then
      Exclude(Result.Mode, dmWriteByOwner);
    Result.MTime := FileTimeToUnixTime(ftLastWriteTime);
    Result.ATime := FileTimeToUnixTime(ftLastAccessTime);
  end;
end;

{ public functions }

procedure KillProcess(Pid: Longint);
var
  h: THandle;
begin
  h := OpenProcess(PROCESS_TERMINATE, False, Pid);
  if h = 0 then
    raise EOSError.Create;
  try
    if not TerminateProcess(h, 0) then
      raise EOSError.Create;
  finally
    CloseHandle(h);
  end;
end;

procedure Remove(const Path: string);
begin
  if IsDir(Path) then
  begin
    if not RemoveDirectory(PChar(Path)) then
      raise EOSError.Create;
  end else
    if not DeleteFile(PChar(Path)) then
      raise EOSError.Create;
end;

procedure CreateDir(const Path: string);
begin
  if not CreateDirectory(PChar(Path), nil) then
    raise EOSError.Create;
end;

procedure ChDir(const Path: string);
begin
  if not SetCurrentDirectory(PChar(Path)) then
    raise EOSError.Create;
end;

function GetWd: string;
begin
  SetString(Result, nil, GetCurrentDirectory(0, nil));
  if GetCurrentDirectory(Length(Result), PChar(Result)) = 0 then
    raise EOSError.Create;
  SetLength(Result, Length(Result)-1); // -1 for null terminate
end;

function SysErrNo: Integer;
begin
  Result := GetLastError;
end;

function SysError(Code: Integer): string;
var
  buf: PChar;
begin
  SetString(Result, buf, FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM
      or FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_IGNORE_INSERTS
      or FORMAT_MESSAGE_MAX_WIDTH_MASK, nil, Code,
      (SUBLANG_DEFAULT shl 10) or LANG_NEUTRAL, PChar(@buf), 0, nil));
  LocalFree(HLOCAL(buf));
  if Result = '' then
    Result := 'Unknown error: ' + IStr(Code);
end;

procedure SetEnv(const Name: string; const Value: string);
begin
  if not SetEnvironmentVariable(PChar(Name), PChar(Value)) then
    raise EOSError.Create;
end;

function GetEnv(const Name: WideString): WideString;
begin
  SetString(Result, nil, GetEnvironmentVariableW(PWideChar(Name), nil, 0));
  GetEnvironmentVariableW(PWideChar(Name), PWideChar(Result), Length(Result));
  SetLength(Result, Length(Result)-1); // -1 for null terminate
end;

function GetEnviron: TStringArray;
var
  i: Integer;
  env, p, q, eq: PChar;
  buf: array[0..1024-1] of Char;
begin
  SetLength(Result, 256);
  i := 0;
  env := GetEnvironmentStrings;
  p := env;
  while p^ <> #0 do
  begin
    q := @buf;
    eq := @buf[High(buf)];
    while (p^ <> #0) and (q < eq) do
    begin
      q^ := p^;
      Inc(q);
      Inc(p);
    end;
    q^ := #0;
    if buf[0] <> '=' then
    begin
      if i = Length(Result) then
        SetLength(Result, i + i div 4);
      SetString(Result[i], buf, q - buf);
      Inc(i);
    end;
    Inc(p);
  end;
  FreeEnvironmentStrings(env);
  SetLength(Result, i);
end;

function Exists(const Path: WideString): Boolean;
begin
  Result := GetFileAttributesW(PWideChar(Path)) <> $FFFFFFFF;
end;

function IsDir(const Path: WideString): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributesW(PWideChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) <> 0);
end;

function IsFile(const Path: WideString): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributesW(PWideChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) = 0);
end;

function DirStat(const Path: WideString): TDir;
var
  finddata: TWin32FindDataW;
  h: THandle;
begin
  // check path exsistence since FindFirstFile accepts wildcard characters
  if GetFileAttributesW(PWideChar(Path)) = $FFFFFFFF then
    raise EIOError.Create(Path);
  h := FindFirstFileW(PWideChar(Path), finddata);
  if h = INVALID_HANDLE_VALUE then
    raise EIOError.Create(Path);
  FindClose(h);
  Result := FindDataToDir(finddata);
end;

procedure DirFWStat(const Path: WideString; const Dir: TDir);
begin
  // TODO: support Mode
  if Dir.Name <> '' then // rename must be last
    if not MoveFileW(PWideChar(Path), PWideChar(Dir.Name)) then
      raise EOSError.Create;
end;

procedure DirWStat(const F: TFile; const Dir: TDir);
var
  ft: TFileTime;
begin
  // TODO: support atime
  with Dir do
  begin
    if MTime <> 0 then
    begin
      ft := UnixTimeToFileTime(MTime);
      if not SetFileTime(F.Handle, nil, nil, @ft) then
        raise EIOError.Create(F.Name);
    end;
    if Mode <> [] then
      if not (dmWriteByOwner in Mode) then
        if not SetFileAttributesW(PWideChar(F.Name),
            FILE_ATTRIBUTE_ARCHIVE or FILE_ATTRIBUTE_READONLY) then
          raise EIOError.Create(F.Name);
  end;
end;

procedure ClearDir(var Dir: TDir);
begin
  with Dir do
  begin
    Name := '';
    Size := 0;
    ATime := 0;
    MTime := 0;
    Mode := [];
  end;
end;

function SameFile(const Path1: WideString; const Path2: WideString): Boolean;
  function FileInfo(const Path: WideString): TByHandleFileInformation;
  var
    h: THandle;
  begin
    h:= CreateFileW(PWideChar(Path), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if h = INVALID_HANDLE_VALUE then
      raise EIOError.Create(Path);
    try
      if not GetFileInformationByHandle(h, Result) then
        raise EIOError.Create(Path);
    finally
      CloseHandle(h);
    end;
  end;
var
  info1, info2: TByHandleFileInformation;
begin
  info1 := FileInfo(Path1);
  info2 := FileInfo(Path2);
  Result := (info1.nFileIndexHigh = info2.nFileIndexHigh) and
      (info1.nFileIndexLow = info2.nFileIndexLow);
end;

function Time: Int64;
var
  ft: TFileTime;
begin
  GetSystemTimeAsFileTime(ft);
  Result := FileTimeToUnixTime(ft);
end;

function NanoSec: Int64;
var
  ft: TFileTime;
begin
  GetSystemTimeAsFileTime(ft);
  Result := (Int64(ft) - UnixTimeDiff) * 100;
end;

procedure SetNanoSec(const NanoSec: Int64);
var
  st: TSystemTime;
begin
  FileTimeToSystemTime(TFileTime(NanoSec div 100 + UnixTimeDiff), st);
  if not SetSystemTime(st) then
    raise EOSError.Create;
end;

function UTCTime(const T: Int64): TTime;
var
  st: TSystemTime;
begin
  FileTimeToSystemTime(UnixTimeToFileTime(T), st);
  with Result do
  begin
    Year := st.wYear;
    Mon := st.wMonth;
    MDay := st.wDay;
    Hour := st.wHour;
    Min := st.wMinute;
    Sec := st.wSecond;
    WDay := st.wDayOfWeek;
    UTCOffset := 0;
  end;
end;

function LocalTime(const T: Int64): TTime;
var
  ut, lt: TFileTime;
  st: TSystemTime;
begin
  ut := UnixTimeToFileTime(T);
  FileTimeToLocalFileTime(ut, lt);
  FileTimeToSystemTime(lt, st);
  with Result do
  begin
    Year := st.wYear;
    Mon := st.wMonth;
    MDay := st.wDay;
    Hour := st.wHour;
    Min := st.wMinute;
    Sec := st.wSecond;
    WDay := st.wDayOfWeek;
    UTCOffset := (Int64(lt) - Int64(ut)) div 10000000;
  end;
end;

procedure Sleep(MilliSecs: Cardinal);
begin
  Windows.Sleep(MilliSecs);
end;

function Clock: Cardinal;
begin
  Result := GetTickCount;
end;

procedure SplitDrive(const Path: WideString; var Drive, Tail: WideString);
begin
  if (Length(Path) > 1) and (Path[2] = DriveSep) then
  begin
    Drive := Copy(Path, 1, 2);
    Tail := Copy(Path, 3, Length(Path)-2);
  end else
  begin
    Drive := '';
    Tail := Path;
  end;
end;

procedure SplitExt(const Path: WideString; var Root, Ext: WideString);
var
  i: Integer;
  head, tail: WideString;
begin
  SplitPath(Path, head, tail);
  i := FindLastSep(tail, ExtSep, ExtSep);
  if i > 0 then
  begin
    Root := Copy(Path, 1, Length(Path)-(Length(tail)-i)-1);
    Ext := Copy(tail, i, Length(tail)-i+1);
  end
  else begin
    Root := Path;
    Ext := '';
  end;
end;

procedure SplitPath(const Path: WideString; var Head, Tail: WideString);
var
  i: Integer;
  d, p: WideString;
begin
  SplitDrive(Path, d, p);
  i := FindLastSep(p, PathSep, AltPathSep);
  if i = 1 then
    Head := d + Copy(p, 1, i)
  else
    Head := d + Copy(p, 1, i-1);
  Tail := Copy(p, i+1, Length(p)-i);
end;

function JoinPath(const Paths: array of WideString): WideString;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to High(Paths) do
  begin
    if Paths[i] = '' then
      Continue;
    if IsAbsPath(Paths[i]) then
      Result := '';
    if (FindLastSep(Paths[i], PathSep, AltPathSep) = Length(Paths[i])) or
        (i = High(Paths)) then
      Result := Result + Paths[i]
    else
      Result := Result + Paths[i] + '\';
  end;
end;

function AbsPath(const Path: WideString): WideString;
var
  p: PWideChar;
  len: DWORD;
begin
  if Path = '' then
  begin
    Result := '';
    Exit;
  end;
  SetString(Result, nil, GetFullPathNameW(PWideChar(Path), 0, nil, p));
  len := GetFullPathNameW(PWideChar(Path), Length(Result), PWideChar(Result), p);
  if len = 0 then
    raise EOSError.Create;
  SetLength(Result, len);
end;

function IsAbsPath(const Path: WideString): Boolean;
var
  drive, tail: WideString;
begin
  SplitDrive(Path, drive, tail);
  Result := (tail <> '') and ((tail[1] = PathSep) or (tail[1] = AltPathSep));
end;

function StrColl(const S1, S2: string): Integer;
begin
  Result := lstrcmp(PChar(S1), PChar(S2));
end;

function StrCollW(const S1, S2: WideString): Integer;
begin
  Result := lstrcmpW(PWideChar(S1), PWideChar(S2));
end;

function StrUpper(const S: string): string;
begin
  SetString(Result, PChar(S), Length(S));
  CharUpperBuff(PChar(Result), Length(Result));
end;

function GenRandom(Len: Longint): string;
var
  prov: HCRYPTPROV;
begin
  if not CryptAcquireContext(prov, nil, nil, PROV_RSA_FULL,
      CRYPT_VERIFYCONTEXT) then
    raise EOSError.Create;
  SetLength(Result, Len);
  try
    if not CryptGenRandom(prov, Len, @Result[1]) then
      raise EOSError.Create;
  finally
    if not CryptReleaseContext(prov, 0) then
      raise EOSError.Create;
  end;
end;


{ TFile }

procedure TFile.CreateHandle(const Name: WideString; Mode: TFileMode;
  Create_: DWORD);
var
  access: DWORD;
begin
  FName := Name;

  if fmRead in Mode then
    access := GENERIC_READ
  else if fmWrite in Mode then
    access := GENERIC_WRITE
  else if fmReadWrite in Mode then
    access := GENERIC_READ or GENERIC_WRITE
  else
    access := 0;

  FHandle := CreateFileW(PWideChar(FName), access,
    FILE_SHARE_READ or FILE_SHARE_WRITE, nil, Create_, FILE_ATTRIBUTE_NORMAL,
    0);
  if FHandle = INVALID_HANDLE_VALUE then
    raise EIOError.Create(FName);
end;

constructor TFile.Create(const Name: WideString; Mode: TFileMode=[fmWrite]);
begin
  CreateHandle(Name, Mode, CREATE_ALWAYS);
end;

constructor TFile.Open(const Name: WideString; Mode: TFileMode=[fmRead]);
begin
  if fmTruncate in Mode then
    CreateHandle(Name, Mode, TRUNCATE_EXISTING)
  else
    CreateHandle(Name, Mode, OPEN_EXISTING);
end;

constructor TFile.CreateStdIn;
begin
  FName := '<STDIN>';
  FHandle := GetStdHandle(STD_INPUT_HANDLE);
  IsStdHandle := True;
end;

constructor TFile.CreateStdOut;
begin
  FName := '<STDOUT>';
  FHandle := GetStdHandle(STD_OUTPUT_HANDLE);
  IsStdHandle := True;
end;

constructor TFile.CreateStdErr;
begin
  FName := '<STDERR>';
  FHandle := GetStdHandle(STD_ERROR_HANDLE);
  IsStdHandle := True;
end;

constructor TFile.OpenConOut;
begin
  Open('CONOUT$', [fmReadWrite]);
end;

constructor TFile.OpenConIn;
begin
  Open('CONIN$', [fmReadWrite]);
end;

constructor TFile.CreateByHandle(Handle: THandle);
begin
  FHandle := Handle;
end;

destructor TFile.Destroy;
begin
  if (FHandle <> INVALID_HANDLE_VALUE) and not IsStdHandle then
    CloseHandle(FHandle);
end;

function TFile.Size: Cardinal;
begin
  Result := GetFileSize(FHandle, nil);
  if Result = $FFFFFFFF then
    raise EIOError.Create(FName);
end;

function TFile.Read(var Buf; NBytes: Cardinal): Cardinal;
begin
  if not ReadFile(FHandle, Buf, DWORD(NBytes), DWORD(Result), nil) then
    if FHandle <> GetStdHandle(STD_INPUT_HANDLE) then
      raise EIOError.Create(FName);
end;

function TFile.Write(const Buf; NBytes: Cardinal): Cardinal;
begin
  if not WriteFile(FHandle, Buf, DWORD(NBytes), DWORD(Result), nil) then
    raise EIOError.Create(FName);
end;

function TFile.Seek(Offset: Int64; Whence: TSeekWhence=swSet): Int64;
var
  li: LARGE_INTEGER;
  method: DWORD;
begin
  case Whence of
    swSet: method := FILE_BEGIN;
    swCur: method := FILE_CURRENT;
    swEnd: method := FILE_END;
  else
    method := FILE_BEGIN;
  end;
  li.QuadPart := Offset;
  li.LowPart := SetFilePointer(FHandle, li.LowPart, @li.HighPart, method);
  if (li.LowPart = $FFFFFFFF) and (GetLastError <> NO_ERROR) then
    raise EIOError.Create(FName);
  Result := li.QuadPart;
end;

function TFile.Seekable: Boolean;
begin
  Result := GetFileType(FHandle) = FILE_TYPE_DISK;
end;

function TFile.IsConsole: Boolean;
begin
  Result := GetFileType(FHandle) = FILE_TYPE_CHAR;
end;


{ TListDir }

constructor TListDir.Create(const Path: WideString);
begin
  Self.Path := Path;
  Handle := FindFirstFileW(PWideChar(JoinPath([Self.Path, '*.*'])), FindData);
  if Handle = INVALID_HANDLE_VALUE then
    raise EIOError.Create(Self.Path);
  First := True;
end;

destructor TListDir.Destroy;
begin
  if Handle <> INVALID_HANDLE_VALUE then
    FindClose(Handle);
  inherited;
end;

function TListDir.Each(var Dir: TDir): Boolean;
begin
  if First then
  begin
    First := False;
    Dir := FindDataToDir(FindData);
    if (Dir.Name <> '.') and (Dir.Name <> '..') then
    begin
      Result := True;
      Exit;
    end;
  end;

  while FindNextFileW(Handle, FindData) do
  begin
    Dir := FindDataToDir(FindData);
    if (Dir.Name <> '.') and (Dir.Name <> '..') then
    begin
      Result := True;
      Exit;
    end;
  end;

  if GetLastError = ERROR_NO_MORE_FILES then
    Result := False
  else
    raise EIOError.Create(Path);
end;

{ TProcess }

constructor TProcess.Create;
var
  arg: WideString;
  si: TStartupInfo;
  inr, outr, outw, outwtmp: THandle;
begin
  arg := JoinCommandLineW(Args);

  FillChar(si, SizeOf(si), 0);
  si.cb := SizeOf(si);
  si.dwFlags := STARTF_USESTDHANDLES;
  si.hStdError := GetStdHandle(STD_ERROR_HANDLE);

  outw := 0;
  if StdOut = nil then
  begin
    CreatePipe(outr, outwtmp, nil, 0);
    DuplicateHandle(GetCurrentProcess, outwtmp, GetCurrentProcess, @outw,
        0, True, DUPLICATE_SAME_ACCESS);
    CloseHandle(outwtmp);
    si.hStdOutput := outw;
    Self.StdOut := TFile.CreateByHandle(outr);
  end
  else begin
    if StdOut.Handle = GetStdHandle(STD_OUTPUT_HANDLE) then
      si.hStdOutput := StdOut.Handle
    else begin
      DuplicateHandle(GetCurrentProcess, StdOut.Handle, GetCurrentProcess,
          @outw, 0, True, DUPLICATE_SAME_ACCESS);
      si.hStdOutput := outw;
    end;
    Self.StdOut := nil;
  end;

  inr := 0;
  if StdIn.Handle = GetStdHandle(STD_INPUT_HANDLE) then
    si.hStdInput := StdIn.Handle
  else begin
    DuplicateHandle(GetCurrentProcess, StdIn.Handle, GetCurrentProcess,
        @inr, 0, True, DUPLICATE_SAME_ACCESS);
    si.hStdInput := inr;
  end;

  try
    if not CreateProcessW(nil, PWideChar(arg), nil, nil, True, 0, nil, nil, si,
        ProcessInformation) then
      raise EOSError.Create;
  finally
    if outw <> 0 then
      CloseHandle(outw);
    if inr <> 0 then
      CloseHandle(inr);
  end;
end;

destructor TProcess.Destroy;
begin
  CloseHandle(ProcessInformation.hProcess);
  CloseHandle(ProcessInformation.hThread);
  StdOut.Free;
end;

function TProcess.Wait;
begin
  WaitForSingleObject(ProcessInformation.hProcess, INFINITE);
  GetExitCodeProcess(ProcessInformation.hProcess, DWORD(Result));
end;

procedure TProcess.Times(var System, User: Int64);
var
  ct, et, kt, ut: FILETIME;
  GetProcessTimes: function (Process: THandle;
      var CreationTime, ExitTime, KernelTime, UserTime: FILETIME): BOOL;
      stdcall;
begin
  GetProcessTimes := GetProcAddress(GetModuleHandle('kernel32.dll'),
      'GetProcessTimes');
  if Assigned(GetProcessTimes) then
  begin
    if not GetProcessTimes(ProcessInformation.hProcess, ct, et, kt, ut) then
      raise EOSError.Create;
    System := Int64(kt) div 10000;
    User := Int64(ut) div 10000;
  end
  else begin
    System := 0;
    User := 0;
  end;
end;

initialization
  Argv := SplitCommandLineW(GetCommandLineW);
  ProgName := Argv[0];
  SetString(ProgPath, nil, MAX_PATH);
  SetLength(ProgPath, GetModuleFileNameW(0, PWideChar(ProgPath), Length(ProgPath)));

end.
